searchNodes=[{"doc":"Bird View Finitomata provides a boilerplate for FSM implementation, allowing to concentrate on the business logic rather than on the process management and transitions/events consistency tweaking. It reads a description of the FSM from a string in PlantUML , Mermaid , or even custom format. Basically, it looks more or less like this PlantUML [ * ] -- &gt; s1 : to_s1 s1 -- &gt; s2 : to_s2 s1 -- &gt; s3 : to_s3 s2 -- &gt; [ * ] : ok s3 -- &gt; [ * ] : ok Mermaid s1 -- &gt; | to_s2 | s2 s1 -- &gt; | to_s3 | s3 Note mermaid does not allow to explicitly specify transitions (and hence event names) from the starting state and to the end state(s), these states names are implicitly set to :* and events to :__start__ and :__end__ respectively. Finitomata validates the FSM is consistent, namely it has a single initial state, one or more final states, and no orphan states. If everything is OK, it generates a GenServer that could be used both alone, and with provided supervision tree. This GenServer requires to implement three callbacks on_transition/4 — mandatory on_failure/3 — optional on_enter/2 — optional on_exit/2 — optional on_terminate/1 — optional on_timer/2 — optional All the callbacks do have a default implementation, that would perfectly handle transitions having a single to state and not requiring any additional business logic attached. Upon start, it moves to the next to initial state and sits there awaiting for the transition request . Then it would call an on_transition/4 callback and move to the next state, or remain in the current one, according to the response. Upon reaching a final state, it would terminate itself. The process keeps all the history of states it went through, and might have a payload in its state. Example Let’s define the FSM instance defmodule MyFSM do @plantuml &quot;&quot;&quot; [*] --&gt; s1 : to_s1 s1 --&gt; s2 : to_s2 s1 --&gt; s3 : to_s3 s2 --&gt; [*] : ok s3 --&gt; [*] : ok &quot;&quot;&quot; use Finitomata , fsm : @plantuml , syntax : Finitomata.PlantUML ## or uncomment lines below for Mermaid syntax # @mermaid &quot;&quot;&quot; # s1 --&gt; |to_s2| s2 # s1 --&gt; |to_s3| s3 # &quot;&quot;&quot; # use Finitomata, fsm: @mermaid, syntax: Finitomata.Mermaid @impl Finitomata def on_transition ( :s1 , :to_s2 , event_payload , state_payload ) , do : { :ok , :s2 , state_payload } end Now we can play with it a bit. children = [ Finitomata . child_spec ( ) ] Supervisor . start_link ( children , strategy : :one_for_one ) Finitomata . start_fsm MyFSM , &quot;My first FSM&quot; , %{ foo : :bar } Finitomata . transition &quot;My first FSM&quot; , { :to_s2 , nil } Finitomata . state &quot;My first FSM&quot; #⇒ %Finitomata.State{current: :s2, history: [:s1], payload: %{foo: :bar}} Finitomata . allowed? &quot;My first FSM&quot; , :* # state #⇒ true Finitomata . responds? &quot;My first FSM&quot; , :to_s2 # event #⇒ false Finitomata . transition &quot;My first FSM&quot; , { :ok , nil } # to final state #⇒ [info] [◉ ⇄] [state: %Finitomata.State{current: :s2, history: [:s1], payload: %{foo: :bar}}] Finitomata . alive? &quot;My first FSM&quot; #⇒ false Typically, one would implement all the on_transition/4 handlers, pattern matching on the state/event. Recurrent Callback If timer: non_neg_integer() option is passed to use Finitomata , then Finitomata.on_timer/2 callback will be executed recurrently. This might be helpful if FSM needs to update its state from the outside world on regular basis. Special Events If the event name is ended with a bang (e. g. idle --&gt; |start!| started ) and this transition is the only one allowed from this state, it’d be considered as determined and FSM will be transitioned into the new state instantly.","ref":"Finitomata.html","title":"Finitomata","type":"behaviour"},{"doc":"Returns true if the supervision tree is alive, false otherwise.","ref":"Finitomata.html#alive?/0","title":"Finitomata.alive?/0","type":"function"},{"doc":"Returns true if the FSM specified is alive, false otherwise.","ref":"Finitomata.html#alive?/1","title":"Finitomata.alive?/1","type":"function"},{"doc":"Returns true if the transition to the state state is possible, false otherwise.","ref":"Finitomata.html#allowed?/2","title":"Finitomata.allowed?/2","type":"function"},{"doc":"This callback will be called on entering the state.","ref":"Finitomata.html#c:on_enter/2","title":"Finitomata.on_enter/2","type":"callback"},{"doc":"This callback will be called on exiting the state.","ref":"Finitomata.html#c:on_exit/2","title":"Finitomata.on_exit/2","type":"callback"},{"doc":"This callback will be called if the transition failed to complete to allow the consumer to take an action upon failure.","ref":"Finitomata.html#c:on_failure/3","title":"Finitomata.on_failure/3","type":"callback"},{"doc":"This callback will be called on transition to the final state to allow the consumer to perform some cleanup, or like.","ref":"Finitomata.html#c:on_terminate/1","title":"Finitomata.on_terminate/1","type":"callback"},{"doc":"This callback will be called recurrently if timer: pos_integer() option has been given to use Finitomata .","ref":"Finitomata.html#c:on_timer/2","title":"Finitomata.on_timer/2","type":"callback"},{"doc":"This callback will be called from each transition processor.","ref":"Finitomata.html#c:on_transition/4","title":"Finitomata.on_transition/4","type":"callback"},{"doc":"Returns true if the transition by the event event is possible, false otherwise.","ref":"Finitomata.html#responds?/2","title":"Finitomata.responds?/2","type":"function"},{"doc":"Starts the FSM instance. The arguments are the implementation of FSM (the module, having use Finitomata ) the name of the FSM (might be any term, but it must be unique) the payload to be carried in the FSM state during the lifecycle The FSM is started supervised.","ref":"Finitomata.html#start_fsm/3","title":"Finitomata.start_fsm/3","type":"function"},{"doc":"The state of the FSM.","ref":"Finitomata.html#state/1","title":"Finitomata.state/1","type":"function"},{"doc":"Initiates the transition. The arguments are the name of the FSM {event, event_payload} tuple; the payload will be passed to the respective on_transition/4 call delay (optional) the interval in milliseconds to apply transition after","ref":"Finitomata.html#transition/3","title":"Finitomata.transition/3","type":"function"},{"doc":"The payload that can be passed to each call to transition/3","ref":"Finitomata.html#t:event_payload/0","title":"Finitomata.event_payload/0","type":"type"},{"doc":"The name of the FSM (might be any term, but it must be unique)","ref":"Finitomata.html#t:fsm_name/0","title":"Finitomata.fsm_name/0","type":"type"},{"doc":"Error types of FSM validation","ref":"Finitomata.html#t:validation_error/0","title":"Finitomata.validation_error/0","type":"type"},{"doc":"Carries the state of the FSM.","ref":"Finitomata.State.html","title":"Finitomata.State","type":"module"},{"doc":"The payload that has been passed to the FSM instance on startup","ref":"Finitomata.State.html#t:payload/0","title":"Finitomata.State.payload/0","type":"type"},{"doc":"The internal representation of the FSM state","ref":"Finitomata.State.html#t:t/0","title":"Finitomata.State.t/0","type":"type"},{"doc":"The internal representation of Transition . It includes from and to states, and the event , all represented as atoms.","ref":"Finitomata.Transition.html","title":"Finitomata.Transition","type":"module"},{"doc":"Returns the list of all the transitions, matching the options. Used internally for the validations. iex&gt; { :ok , transitions } = ...&gt; Finitomata.Mermaid . parse ( ...&gt; &quot;idle --&gt; |to_s1| s1 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s2| s2 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s3| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |to_s3| s3&quot; ) ...&gt; Finitomata.Transition . allowed ( transitions , to : [ :idle , :* ] ) [ { :* , :idle , :__start__ } , { :s3 , :* , :__end__ } ] iex&gt; Finitomata.Transition . allowed ( transitions , from : :s1 ) [ { :s1 , :s2 , :to_s2 } , { :s1 , :s3 , :to_s3 } ] iex&gt; Finitomata.Transition . allowed ( transitions , from : :s1 , to : :s3 ) [ { :s1 , :s3 , :to_s3 } ] iex&gt; Finitomata.Transition . allowed ( transitions , from : :s1 , with : :to_s3 ) [ { :s1 , :s3 , :to_s3 } ] iex&gt; Finitomata.Transition . allowed ( transitions , from : :s2 , with : :to_s2 ) [ ]","ref":"Finitomata.Transition.html#allowed/2","title":"Finitomata.Transition.allowed/2","type":"function"},{"doc":"Returns the list of all the transitions, matching the from state and the event . Used internally for the validations. iex&gt; { :ok , transitions } = ...&gt; Finitomata.PlantUML . parse ( &quot;[*] --&gt; s1 : foo \\n s1 --&gt; s2 : ok \\n s2 --&gt; [*] : ko&quot; ) ...&gt; Finitomata.Transition . allowed ( transitions , :s1 , :foo ) [ :s2 ] ...&gt; Finitomata.Transition . allowed ( transitions , :s1 , :* ) [ ]","ref":"Finitomata.Transition.html#allowed/3","title":"Finitomata.Transition.allowed/3","type":"function"},{"doc":"Returns true if the transition from → to is allowed, false otherwise. iex&gt; { :ok , transitions } = ...&gt; Finitomata.PlantUML . parse ( &quot;[*] --&gt; s1 : foo \\n s1 --&gt; s2 : ok \\n s2 --&gt; [*] : ko&quot; ) ...&gt; Finitomata.Transition . allowed? ( transitions , :s1 , :s2 ) true ...&gt; Finitomata.Transition . allowed? ( transitions , :s1 , :* ) false","ref":"Finitomata.Transition.html#allowed?/3","title":"Finitomata.Transition.allowed?/3","type":"function"},{"doc":"Returns Finitomata.Transition.event() if there is a determined transition from the current state. The transition is determined, if it is the only transition allowed from the state. Used internally for the validations. iex&gt; { :ok , transitions } = ...&gt; Finitomata.Mermaid . parse ( ...&gt; &quot;idle --&gt; |to_s1| s1 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s2| s2 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s3| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |to_s3| s3&quot; ) ...&gt; Finitomata.Transition . determined ( transitions ) [ s3 : { :__end__ , :* } , s2 : { :to_s3 , :s3 } , idle : { :to_s1 , :s1 } ]","ref":"Finitomata.Transition.html#determined/1","title":"Finitomata.Transition.determined/1","type":"function"},{"doc":"Returns {:ok, {event(), state()}} tuple if there is a determined transition from the current state, :error otherwise. The transition is determined, if it is the only transition allowed from the state. Used internally for the validations. iex&gt; { :ok , transitions } = ...&gt; Finitomata.Mermaid . parse ( ...&gt; &quot;idle --&gt; |to_s1| s1 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s2| s2 \\n &quot; &lt;&gt; ...&gt; &quot;s1 --&gt; |to_s3| s3 \\n &quot; &lt;&gt; ...&gt; &quot;s2 --&gt; |to_s3| s3&quot; ) ...&gt; Finitomata.Transition . determined ( transitions , :s1 ) :error iex&gt; Finitomata.Transition . determined ( transitions , :s2 ) { :ok , { :to_s3 , :s3 } } iex&gt; Finitomata.Transition . determined ( transitions , :s3 ) { :ok , { :__end__ , :* } }","ref":"Finitomata.Transition.html#determined/2","title":"Finitomata.Transition.determined/2","type":"function"},{"doc":"Returns the state after starting one, so-called entry state. iex&gt; { :ok , transitions } = ...&gt; Finitomata.PlantUML . parse ( &quot;[*] --&gt; s1 : foo \\n s1 --&gt; s2 : ok \\n s2 --&gt; [*] : ko&quot; ) ...&gt; Finitomata.Transition . entry ( transitions ) :s1","ref":"Finitomata.Transition.html#entry/1","title":"Finitomata.Transition.entry/1","type":"function"},{"doc":"Returns true if the state from hsa an outgoing transition with event , false otherwise. iex&gt; { :ok , transitions } = ...&gt; Finitomata.PlantUML . parse ( &quot;[*] --&gt; s1 : foo \\n s1 --&gt; s2 : ok \\n s2 --&gt; [*] : ko&quot; ) ...&gt; Finitomata.Transition . responds? ( transitions , :s1 , :ok ) true ...&gt; Finitomata.Transition . responds? ( transitions , :s1 , :ko ) false","ref":"Finitomata.Transition.html#responds?/3","title":"Finitomata.Transition.responds?/3","type":"function"},{"doc":"Returns the not ordered list of states, excluding the starting and ending states :* . iex&gt; { :ok , transitions } = ...&gt; Finitomata.PlantUML . parse ( &quot;[*] --&gt; s1 : foo \\n s1 --&gt; s2 : ok \\n s2 --&gt; [*] : ko&quot; ) ...&gt; Finitomata.Transition . states ( transitions ) [ :s1 , :s2 ]","ref":"Finitomata.Transition.html#states/1","title":"Finitomata.Transition.states/1","type":"function"},{"doc":"The event in FSM","ref":"Finitomata.Transition.html#t:event/0","title":"Finitomata.Transition.event/0","type":"type"},{"doc":"The state of FSM","ref":"Finitomata.Transition.html#t:state/0","title":"Finitomata.Transition.state/0","type":"type"},{"doc":"The transition is represented by from and to states and the event .","ref":"Finitomata.Transition.html#t:t/0","title":"Finitomata.Transition.t/0","type":"type"},{"doc":"The FSM boilerplate based on callbacks","ref":"readme.html","title":"Finitomata","type":"extras"},{"doc":"Bird View Finitomata provides a boilerplate for FSM implementation, allowing to concentrate on the business logic rather than on the process management and transitions/events consistency tweaking. It reads a description of the FSM from a string in PlantUML , Mermaid , or even custom format. Basically, it looks more or less like this PlantUML [ * ] -- &gt; s1 : to_s1 s1 -- &gt; s2 : to_s2 s1 -- &gt; s3 : to_s3 s2 -- &gt; [ * ] : ok s3 -- &gt; [ * ] : ok Mermaid s1 -- &gt; | to_s2 | s2 s1 -- &gt; | to_s3 | s3 Note mermaid does not allow to explicitly specify transitions (and hence event names) from the starting state and to the end state(s), these states names are implicitly set to :* and events to :__start__ and :__end__ respectively. Finitomata validates the FSM is consistent, namely it has a single initial state, one or more final states, and no orphan states. If everything is OK, it generates a GenServer that could be used both alone, and with provided supervision tree. This GenServer requires to implement three callbacks on_transition/4 — mandatory on_failure/3 — optional on_enter/2 — optional on_exit/2 — optional on_terminate/1 — optional on_timer/2 — optional All the callbacks do have a default implementation, that would perfectly handle transitions having a single to state and not requiring any additional business logic attached. Upon start, it moves to the next to initial state and sits there awaiting for the transition request . Then it would call an on_transition/4 callback and move to the next state, or remain in the current one, according to the response. Upon reaching a final state, it would terminate itself. The process keeps all the history of states it went through, and might have a payload in its state. Example Let’s define the FSM instance defmodule MyFSM do @plantuml &quot;&quot;&quot; [*] --&gt; s1 : to_s1 s1 --&gt; s2 : to_s2 s1 --&gt; s3 : to_s3 s2 --&gt; [*] : ok s3 --&gt; [*] : ok &quot;&quot;&quot; use Finitomata , fsm : @plantuml , syntax : Finitomata.PlantUML ## or uncomment lines below for Mermaid syntax # @mermaid &quot;&quot;&quot; # s1 --&gt; |to_s2| s2 # s1 --&gt; |to_s3| s3 # &quot;&quot;&quot; # use Finitomata, fsm: @mermaid, syntax: Finitomata.Mermaid @impl Finitomata def on_transition ( :s1 , :to_s2 , event_payload , state_payload ) , do : { :ok , :s2 , state_payload } end Now we can play with it a bit. children = [ Finitomata . child_spec ( ) ] Supervisor . start_link ( children , strategy : :one_for_one ) Finitomata . start_fsm MyFSM , &quot;My first FSM&quot; , %{ foo : :bar } Finitomata . transition &quot;My first FSM&quot; , { :to_s2 , nil } Finitomata . state &quot;My first FSM&quot; #⇒ %Finitomata.State{current: :s2, history: [:s1], payload: %{foo: :bar}} Finitomata . allowed? &quot;My first FSM&quot; , :* # state #⇒ true Finitomata . responds? &quot;My first FSM&quot; , :to_s2 # event #⇒ false Finitomata . transition &quot;My first FSM&quot; , { :ok , nil } # to final state #⇒ [info] [◉ ⇄] [state: %Finitomata.State{current: :s2, history: [:s1], payload: %{foo: :bar}}] Finitomata . alive? &quot;My first FSM&quot; #⇒ false Typically, one would implement all the on_transition/4 handlers, pattern matching on the state/event. Recurrent Callback If timer: non_neg_integer() option is passed to use Finitomata , then Finitomata.on_timer/2 callback will be executed recurrently. This might be helpful if FSM needs to update its state from the outside world on regular basis. Special Events If the event name is ended with a bang (e. g. idle --&gt; |start!| started ) and this transition is the only one allowed from this state, it’d be considered as determined and FSM will be transitioned into the new state instantly.","ref":"readme.html#introduction","title":"Finitomata - Introduction","type":"extras"},{"doc":"def deps do [ { :finitomata , &quot;~&gt; 0.1&quot; } ] end","ref":"readme.html#installation","title":"Finitomata - Installation","type":"extras"},{"doc":"0.6.3 — soft? events which do not call on_failure/2 and do not log errors 0.6.2 — ensure_entry: option to retry a transition 0.6.1 — code cleanup + auto_terminate: option to make :__end__ transition imminent 0.6.0 — on_timer/2 and banged imminent transitions 0.5.2 — state() type on generated FSMs 0.5.1 — fixed specs [credits @egidijusz] 0.5.0 — all callbacks but on_transition/4 are optional, accept impl_for: param to use Finitomata 0.4.0 — allow anonymous FSM instances 0.3.0 — en_entry/2 and on_exit/2 optional callbacks 0.2.0 — Mermaid support Documentation .","ref":"readme.html#changelog","title":"Finitomata - Changelog","type":"extras"}]